/*
* Copyright (c) 2008-2017, NVIDIA CORPORATION. All rights reserved.
*
* NVIDIA CORPORATION and its licensors retain all intellectual property
* and proprietary rights in and to this software, related documentation
* and any modifications thereto. Any use, reproduction, disclosure or
* distribution of this software and related documentation without an express
* license agreement from NVIDIA CORPORATION is strictly prohibited.
*/


#include "./sky_renderer.h"
#include "./test_util.h"

#if USE_VK
	#include <shaders/bytecode/SkyRenderingVS_vk.h>
	#include <shaders/bytecode/SkyRenderingPS_vk.h>
	#include <shaders/bytecode/FullscreenVS_vk.h>
	#include <shaders/bytecode/SkyUpdatePS_vk.h>
	#include <shaders/bytecode/MipmapGenerationPS_vk.h>
#else
	#include <shaders/bytecode/SkyRenderingVS_dx.h>
	#include <shaders/bytecode/SkyRenderingPS_dx.h>
	#include <shaders/bytecode/FullscreenVS_dx.h>
	#include <shaders/bytecode/SkyUpdatePS_dx.h>
	#include <shaders/bytecode/MipmapGenerationPS_dx.h>
#endif // USE_VK

struct vs_sky_cbuffer
{
	gfsdk_float4x4 g_matViewProj;
};

struct ps_sky_cbuffer
{
	gfsdk_float3 g_sunDirection;
	float g_sunIntensity;
	float g_skyRotation;
	float g_rayleighScattering;
	float g_mieScattering;
	float g_useDynamicSkyTexture;
};

#ifndef PI 
#define PI 3.141593f
#endif

bool SkyRenderer::Init()
{
	nvrhi::CommandListParameters cmdListParams;
	m_hCommandList = m_pDevice->createCommandList(cmdListParams);
	m_hCommandList->open();

	// Create depth buffer 
	// It is created already if BackBufferResized was called, so destroying it first then
	if (m_pState->hDepthBufferTexture != nullptr)
	{
		m_pState->hDepthBufferTexture = nullptr;
	}
	nvrhi::TextureDesc td;
	td.width = m_backBufferWidth;
	td.height = m_backBufferHeight;
	td.sampleCount = m_sampleCount;
	td.debugName = "Depth target created in SkyRenderer::Init()";
	td.format = nvrhi::Format::D24S8;
	td.isRenderTarget = true;
	td.clearValue = nvrhi::Color(1.0f);
	td.useClearValue = true;
	td.isTypeless = true; // Depth targets must be declared typeless in NVRHI
	m_pState->hDepthBufferTexture = m_pDevice->createTexture(td);
	m_hCommandList->beginTrackingTextureState(m_pState->hDepthBufferTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::COMMON);
	m_hCommandList->beginTrackingTextureState(m_pState->hDepthBufferTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::DEPTH_WRITE);

	// Initialize frame buffer #0
	td.debugName = "proxy color target";
#if USE_VK
	td.format = nvrhi::Format::BGRA8_UNORM;
#else
	td.format = nvrhi::Format::RGBA8_UNORM;
#endif
	m_hInitTimeRenderTargetTexture = m_pDevice->createTexture(td);
	nvrhi::FramebufferDesc fbDesc;
	fbDesc.addColorAttachment(m_hInitTimeRenderTargetTexture, 0, 0);
	fbDesc.setDepthAttachment(m_pState->hDepthBufferTexture, 0, 0);
	m_hInitTimeFB = m_pDevice->createFramebuffer(fbDesc);


	// Load shaders
#if USE_VK
	auto VSBinary = SkyRenderingVS_bytecode_vk;
	auto VSBinarySize = sizeof(SkyRenderingVS_bytecode_vk);
	auto PSBinary = SkyRenderingPS_bytecode_vk;
	auto PSBinarySize = sizeof(SkyRenderingPS_bytecode_vk);
#else 
	auto VSBinary = SkyRenderingVS_bytecode_dx;
	auto VSBinarySize = sizeof(SkyRenderingVS_bytecode_dx);
	auto PSBinary = SkyRenderingPS_bytecode_dx;
	auto PSBinarySize = sizeof(SkyRenderingPS_bytecode_dx);
#endif 

	m_hRenderSkyVS = m_pDevice->createShader(nvrhi::ShaderDesc(nvrhi::ShaderType::SHADER_VERTEX), VSBinary, VSBinarySize);
	if (m_hRenderSkyVS == nullptr) return false;

	m_hRenderSkyPS = m_pDevice->createShader(nvrhi::ShaderDesc(nvrhi::ShaderType::SHADER_PIXEL), PSBinary, PSBinarySize);
	if (m_hRenderSkyPS == nullptr) return false;

	// Create input layout
	nvrhi::VertexAttributeDesc render_sky_layout[] = {
		{ "POSITION", nvrhi::Format::RGBA32_FLOAT, 1, 0, 0, 24, false },
		{ "TEXCOORD", nvrhi::Format::RG32_FLOAT, 1, 0, 16, 24, false },
	};

	m_hRenderSkyLayout = m_pDevice->createInputLayout(render_sky_layout, 2, m_hRenderSkyVS);

	// Create textures
	m_hStaticSkyDomeTexture = CreateTextureFromFile(m_pDevice, m_hCommandList, (GetMediaFolder()/"sky_dome_1.dds").generic_string().c_str());

	if (m_hStaticSkyDomeTexture == nullptr) return false;

	td.width = SKYDOME_TEXTURE_SIZE;
	td.height = SKYDOME_TEXTURE_SIZE;
	td.sampleCount = 1;
	td.debugName = "Dynamic sky texture";
	td.format = nvrhi::Format::RGBA16_FLOAT;
	td.isRenderTarget = true;
	td.mipLevels = SKYDOME_TEXTURE_MIPMAPCOUNT;
	m_pState->hDynamicSkyDomeTexture = m_pDevice->createTexture(td);

	// Create samplers
	nvrhi::SamplerDesc samplerDesc;
	samplerDesc.magFilter = true;
	samplerDesc.minFilter = true;
	samplerDesc.mipFilter = true;
	samplerDesc.wrapMode[0] = nvrhi::SamplerDesc::WrapMode::WRAP_MODE_CLAMP;
	samplerDesc.wrapMode[1] = nvrhi::SamplerDesc::WrapMode::WRAP_MODE_CLAMP; // Need it to avoid issues sampling the skydome texture in singularities in zenith and nadir points
	samplerDesc.wrapMode[2] = nvrhi::SamplerDesc::WrapMode::WRAP_MODE_WRAP;
	m_hBilinearSampler = m_pDevice->createSampler(samplerDesc);
	samplerDesc.anisotropy = 16.0f;
	m_hAnisotropicSampler = m_pDevice->createSampler(samplerDesc);

	// Init constant buffer for vertex shader
	nvrhi::BufferDesc cbDesc;
	cbDesc.byteSize = sizeof(vs_sky_cbuffer);
	cbDesc.debugName = "Sky VS CB";
	cbDesc.isConstantBuffer = true;
	m_hRenderSkyVSCB = m_pDevice->createBuffer(cbDesc);

	// Init constant buffer for pixel shader
	cbDesc.byteSize = sizeof(ps_sky_cbuffer);
	cbDesc.debugName = "Sky PS CB";
	cbDesc.isConstantBuffer = true;
	m_hPSCB = m_pDevice->createBuffer(cbDesc);

	// Create sky rendering PSO & bindings
	nvrhi::BlendState blendState;
	blendState.blendEnable[0] = false;

	nvrhi::RasterState rasterState;
	rasterState.fillMode = nvrhi::RasterState::FILL_SOLID;
	rasterState.cullMode = nvrhi::RasterState::CULL_NONE;
	rasterState.scissorEnable = false;
	rasterState.depthClipEnable = false;

	nvrhi::DepthStencilState depthStencilState;
	depthStencilState.depthEnable = true;
	depthStencilState.depthWriteMask = nvrhi::DepthStencilState::DEPTH_WRITE_MASK_ALL;
	depthStencilState.depthFunc = nvrhi::DepthStencilState::COMPARISON_LESS_EQUAL;
	depthStencilState.stencilEnable = false;

	nvrhi::RenderState renderState;
	renderState.blendState = blendState;
	renderState.depthStencilState = depthStencilState;
	renderState.rasterState = rasterState;

	nvrhi::GraphicsPipelineDesc psoDesc;
	nvrhi::BindingSetDesc bsDesc;
	nvrhi::BindingLayoutDesc layoutDesc;

	layoutDesc.VS = { { 0, nvrhi::ResourceType::ConstantBuffer, 0 } };
	layoutDesc.PS = { { 0, nvrhi::ResourceType::Texture_SRV, 0},
						{ 0, nvrhi::ResourceType::Sampler, 0 },
						{ 0, nvrhi::ResourceType::ConstantBuffer, 0 } };
	m_hRenderSkyBindingLayout =  m_pDevice->createBindingLayout(layoutDesc);

	bsDesc.VS = { nvrhi::BindingSetItem::ConstantBuffer(0, m_hRenderSkyVSCB ) };
	bsDesc.PS = { nvrhi::BindingSetItem::Texture_SRV(0,m_pState->hDynamicSkyDomeTexture, nvrhi::Format::RGBA16_FLOAT ),
					nvrhi::BindingSetItem::Sampler(0, m_hBilinearSampler ),
					nvrhi::BindingSetItem::ConstantBuffer(0, m_hPSCB ) };
	m_hRenderSkyBindingSet = m_pDevice->createBindingSet(bsDesc, m_hRenderSkyBindingLayout);

	psoDesc.bindingLayouts = { m_hRenderSkyBindingLayout };
	psoDesc.VS = m_hRenderSkyVS;
	psoDesc.PS = m_hRenderSkyPS;
	psoDesc.primType = nvrhi::PrimitiveType::TRIANGLE_STRIP;
	psoDesc.inputLayout = m_hRenderSkyLayout;
	psoDesc.renderState = renderState;
	m_hRenderSkyPSO = m_pDevice->createGraphicsPipeline(psoDesc, m_hInitTimeFB);



	// Create VB and IB for skydome
	uint32_t skydome_vertices_count = SKYDOME_SLICES_COUNT*(SKYDOME_SLICES_COUNT + 2) * 2;
	
	float *skydome_vertexdata = (float*) malloc(skydome_vertices_count * 6 * sizeof(float));
	uint32_t *skydome_indexdata = (uint32_t*) malloc(skydome_vertices_count * sizeof(uint32_t));

	for (uint32_t i = 0; i < skydome_vertices_count; i++) skydome_indexdata[i] = i;
	
	float skydome_radius = m_pState->fCameraClipFar*0.8f;

	for (int j = 0; j < SKYDOME_SLICES_COUNT; j++)
	{
		int i = 0;
		int floatnum = (j*(SKYDOME_SLICES_COUNT + 2) * 2) * 6;
		skydome_vertexdata[floatnum + 0] = skydome_radius*cos(2.0f*PI*(float)i / (float)SKYDOME_SLICES_COUNT)*cos(-0.5f*PI + PI*(float)j / (float)SKYDOME_SLICES_COUNT);
		skydome_vertexdata[floatnum + 2] = skydome_radius*sin(-0.5f*PI + PI*(float)(j) / (float)SKYDOME_SLICES_COUNT);
		skydome_vertexdata[floatnum + 1] = skydome_radius*sin(2.0f*PI*(float)i / (float)SKYDOME_SLICES_COUNT)*cos(-0.5f*PI + PI*(float)j / (float)SKYDOME_SLICES_COUNT);
		skydome_vertexdata[floatnum + 3] = 1.0f;
		skydome_vertexdata[floatnum + 4] = (float)i / (float)SKYDOME_SLICES_COUNT;
		skydome_vertexdata[floatnum + 5] = (float)(j + 0) / (float)SKYDOME_SLICES_COUNT;
		floatnum += 6;
		for (i = 0; i < SKYDOME_SLICES_COUNT + 1; i++)
		{
			skydome_vertexdata[floatnum + 0] = skydome_radius*cos(2.0f*PI*(float)i / (float)SKYDOME_SLICES_COUNT)*cos(-0.5f*PI + PI*(float)j / (float)SKYDOME_SLICES_COUNT);
			skydome_vertexdata[floatnum + 2] = skydome_radius*sin(-0.5f*PI + PI*(float)(j) / (float)SKYDOME_SLICES_COUNT);
			skydome_vertexdata[floatnum + 1] = skydome_radius*sin(2.0f*PI*(float)i / (float)SKYDOME_SLICES_COUNT)*cos(-0.5f*PI + PI*(float)j / (float)SKYDOME_SLICES_COUNT);
			skydome_vertexdata[floatnum + 3] = 1.0f;
			skydome_vertexdata[floatnum + 4] = (float)i / (float)SKYDOME_SLICES_COUNT;
			skydome_vertexdata[floatnum + 5] = 1.0f*(float)(j + 0) / (float)SKYDOME_SLICES_COUNT;
			floatnum += 6;
			skydome_vertexdata[floatnum + 0] = skydome_radius*cos(2.0f*PI*(float)i / (float)SKYDOME_SLICES_COUNT)*cos(-0.5f*PI + PI*(float)(j + 1) / (float)SKYDOME_SLICES_COUNT);
			skydome_vertexdata[floatnum + 2] = skydome_radius*sin(-0.5f*PI + PI*(float)(j + 1) / (float)SKYDOME_SLICES_COUNT);
			skydome_vertexdata[floatnum + 1] = skydome_radius*sin(2.0f*PI*(float)i / (float)SKYDOME_SLICES_COUNT)*cos(-0.5f*PI + PI*(float)(j + 1) / (float)SKYDOME_SLICES_COUNT);
			skydome_vertexdata[floatnum + 3] = 1.0f;
			skydome_vertexdata[floatnum + 4] = (float)i / (float)SKYDOME_SLICES_COUNT;
			skydome_vertexdata[floatnum + 5] = 1.0f*(float)(j + 1) / (float)SKYDOME_SLICES_COUNT;
			floatnum += 6;
		}
		i = SKYDOME_SLICES_COUNT;
		skydome_vertexdata[floatnum + 0] = skydome_radius*cos(2.0f*PI*(float)i / (float)SKYDOME_SLICES_COUNT)*cos(-0.5f*PI + PI*(float)(j + 1) / (float)SKYDOME_SLICES_COUNT);
		skydome_vertexdata[floatnum + 2] = skydome_radius*sin(-0.5f*PI + PI*(float)(j + 1) / (float)SKYDOME_SLICES_COUNT);
		skydome_vertexdata[floatnum + 1] = skydome_radius*sin(2.0f*PI*(float)i / (float)SKYDOME_SLICES_COUNT)*cos(-0.5f*PI + PI*(float)(j + 1) / (float)SKYDOME_SLICES_COUNT);
		skydome_vertexdata[floatnum + 3] = 1.0f;
		skydome_vertexdata[floatnum + 4] = (float)i / (float)SKYDOME_SLICES_COUNT;
		skydome_vertexdata[floatnum + 5] = 1.0f*(float)(j + 1) / (float)SKYDOME_SLICES_COUNT;
	}

	nvrhi::BufferDesc bd;
	bd.byteSize = skydome_vertices_count * 6 * sizeof(float);
	bd.isVertexBuffer = true;
	bd.isIndexBuffer = false;
	m_hRenderSkyVertexBuffer = m_pDevice->createBuffer(bd);
	m_hCommandList->beginTrackingBufferState(m_hRenderSkyVertexBuffer, nvrhi::ResourceStates::COMMON);
	m_hCommandList->writeBuffer(m_hRenderSkyVertexBuffer, skydome_vertexdata, bd.byteSize);
	m_hCommandList->endTrackingBufferState(m_hRenderSkyVertexBuffer, nvrhi::ResourceStates::VERTEX_BUFFER, true);
	
	bd.byteSize = skydome_vertices_count * sizeof(uint32_t);
	bd.isVertexBuffer = false;
	bd.isIndexBuffer = true;
	m_hRenderSkyIndexBuffer = m_pDevice->createBuffer(bd);
	m_hCommandList->beginTrackingBufferState(m_hRenderSkyIndexBuffer, nvrhi::ResourceStates::COMMON);
	m_hCommandList->writeBuffer(m_hRenderSkyIndexBuffer, skydome_indexdata, bd.byteSize);
	m_hCommandList->endTrackingBufferState(m_hRenderSkyIndexBuffer, nvrhi::ResourceStates::INDEX_BUFFER, true);

	free(skydome_vertexdata);
	free(skydome_indexdata);

	// Creating resources for updating sky
	
	// Shaders
#if USE_VK
	VSBinary = FullscreenVS_bytecode_vk;
	VSBinarySize = sizeof(FullscreenVS_bytecode_vk);
	PSBinary = SkyUpdatePS_bytecode_vk;
	PSBinarySize = sizeof(SkyUpdatePS_bytecode_vk);
#else 
	VSBinary = FullscreenVS_bytecode_dx;
	VSBinarySize = sizeof(FullscreenVS_bytecode_dx);
	PSBinary = SkyUpdatePS_bytecode_dx;
	PSBinarySize = sizeof(SkyUpdatePS_bytecode_dx);
#endif 

	m_hFullscreenVS = m_pDevice->createShader(nvrhi::ShaderDesc(nvrhi::ShaderType::SHADER_VERTEX), VSBinary, VSBinarySize);
	if (m_hFullscreenVS == nullptr) return false;

	m_hUpdateSkyPS = m_pDevice->createShader(nvrhi::ShaderDesc(nvrhi::ShaderType::SHADER_PIXEL), PSBinary, PSBinarySize);
	if (m_hUpdateSkyPS == nullptr) return false;

	// Input layout
	nvrhi::VertexAttributeDesc update_sky_layout[] = {
		{ "POSITION", nvrhi::Format::RGBA32_FLOAT, 1, 0, 0, 24, false },
		{ "TEXCOORD", nvrhi::Format::RG32_FLOAT, 1, 0, 16, 24, false },
	};
	m_hUpdateSkyLayout = m_pDevice->createInputLayout(update_sky_layout, 2, m_hFullscreenVS);

	// Draw state
	depthStencilState.depthEnable = false;
	renderState.blendState = blendState;
	renderState.depthStencilState = depthStencilState;
	renderState.rasterState = rasterState;

	// PSO & bindings
	layoutDesc.VS = { };
	layoutDesc.PS = { { 0, nvrhi::ResourceType::Texture_SRV, 0 },
						{ 0, nvrhi::ResourceType::Sampler, 0 },
						{ 0, nvrhi::ResourceType::ConstantBuffer, 0 } };
	m_hUpdateSkyBindingLayout = m_pDevice->createBindingLayout(layoutDesc);

	bsDesc.VS = { };
	bsDesc.PS = { nvrhi::BindingSetItem::Texture_SRV(0, m_hStaticSkyDomeTexture, nvrhi::Format::RGBA8_UNORM ),
					nvrhi::BindingSetItem::Sampler(0, m_hBilinearSampler ),
					nvrhi::BindingSetItem::ConstantBuffer(0, m_hPSCB ) };
	m_hUpdateSkyBindingSet = m_pDevice->createBindingSet(bsDesc, m_hUpdateSkyBindingLayout);
	
	fbDesc.colorAttachments[0] = { m_pState->hDynamicSkyDomeTexture, 0, 0 };
	fbDesc.depthAttachment = nullptr;
	m_hUpdateSkyFB = m_pDevice->createFramebuffer(fbDesc);

	psoDesc.bindingLayouts = { m_hUpdateSkyBindingLayout };
	psoDesc.VS = m_hFullscreenVS;
	psoDesc.PS = m_hUpdateSkyPS;
	psoDesc.primType = nvrhi::PrimitiveType::TRIANGLE_STRIP;
	psoDesc.inputLayout = m_hUpdateSkyLayout;
	psoDesc.renderState = renderState;
	m_hUpdateSkyPSO = m_pDevice->createGraphicsPipeline(psoDesc, m_hUpdateSkyFB);

	// Index & vertex buffers
	float tex_adjust = 0.0f;
	float vertices[24] = { -1.0f,  1.0f, 0,	1.0f, tex_adjust,        tex_adjust,
							-1.0f, -1.0f, 0, 1.0f, tex_adjust,        tex_adjust + 1.0f,
							1.0f,  1.0f, 0, 1.0f, tex_adjust + 1.0f, tex_adjust,
							1.0f, -1.0f, 0, 1.0f, tex_adjust + 1.0f, tex_adjust + 1.0f };
	uint32_t indices[4] = { 0, 1, 2, 3 };

	bd.byteSize = 24 * sizeof(float);
	bd.isVertexBuffer = true;
	bd.isIndexBuffer = false;
	m_hUpdateSkyVertexBuffer = m_pDevice->createBuffer(bd);
	m_hCommandList->beginTrackingBufferState(m_hUpdateSkyVertexBuffer, nvrhi::ResourceStates::COMMON);
	m_hCommandList->writeBuffer(m_hUpdateSkyVertexBuffer, vertices, bd.byteSize);
	m_hCommandList->endTrackingBufferState(m_hUpdateSkyVertexBuffer, nvrhi::ResourceStates::VERTEX_BUFFER, true);


	bd.byteSize = 4 * sizeof(uint32_t);
	bd.isVertexBuffer = false;
	bd.isIndexBuffer = true;
	m_hUpdateSkyIndexBuffer = m_pDevice->createBuffer(bd);
	m_hCommandList->beginTrackingBufferState(m_hUpdateSkyIndexBuffer, nvrhi::ResourceStates::COMMON);
	m_hCommandList->writeBuffer(m_hUpdateSkyIndexBuffer, indices, bd.byteSize);
	m_hCommandList->endTrackingBufferState(m_hUpdateSkyIndexBuffer, nvrhi::ResourceStates::INDEX_BUFFER, true);

	// Mipmap generation shader
#if USE_VK
	PSBinary = MipmapGenerationPS_bytecode_vk;
	PSBinarySize = sizeof(MipmapGenerationPS_bytecode_vk);
#else 
	PSBinary = MipmapGenerationPS_bytecode_dx;
	PSBinarySize = sizeof(MipmapGenerationPS_bytecode_dx);
#endif 

	m_hMipGenPS = m_pDevice->createShader(nvrhi::ShaderDesc(nvrhi::ShaderType::SHADER_PIXEL), PSBinary, PSBinarySize);
	if (m_hMipGenPS == nullptr) return false;

	// Mipmap generation PSO & bindings

	layoutDesc.VS = {};
	layoutDesc.PS = { { 0, nvrhi::ResourceType::Texture_SRV, 0 },
						{ 0, nvrhi::ResourceType::Sampler, 0 } };
	m_hMipGenBindingLayout = m_pDevice->createBindingLayout(layoutDesc);

	psoDesc.bindingLayouts = { m_hMipGenBindingLayout };
	psoDesc.VS = m_hFullscreenVS;
	psoDesc.PS = m_hMipGenPS;

	int target_mip_size = SKYDOME_TEXTURE_SIZE / 2;
	for (int mip_level = 0; mip_level < SKYDOME_TEXTURE_MIPMAPCOUNT - 1; mip_level++)
	{
		fbDesc.colorAttachments[0].texture = m_pState->hDynamicSkyDomeTexture;
		fbDesc.colorAttachments[0].subresources = nvrhi::TextureSubresourceSet(mip_level + 1, 1, 0, 1);
		m_hMipGenFB[mip_level] = m_pDevice->createFramebuffer(fbDesc);
		m_hMipGenPSO[mip_level] = m_pDevice->createGraphicsPipeline(psoDesc, m_hMipGenFB[mip_level]);

		bsDesc.VS = {};
		bsDesc.PS = { nvrhi::BindingSetItem::Texture_SRV(0, m_pState->hDynamicSkyDomeTexture,  nvrhi::Format::RGBA16_FLOAT, nvrhi::TextureSubresourceSet(mip_level, 1, 0, 1) ),
						nvrhi::BindingSetItem::Sampler(0, m_hBilinearSampler ) };
		m_hMipGenBindingSet[mip_level] = m_pDevice->createBindingSet(bsDesc, m_hMipGenBindingLayout);

		target_mip_size /= 2;
	}

	// Set initial state for resources
	m_hCommandList->beginTrackingTextureState(m_pState->hDynamicSkyDomeTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::COMMON);
	m_hCommandList->beginTrackingBufferState(m_hRenderSkyVSCB, nvrhi::ResourceStates::COMMON);
	m_hCommandList->beginTrackingBufferState(m_hPSCB, nvrhi::ResourceStates::COMMON);

	m_hCommandList->endTrackingTextureState(m_pState->hDynamicSkyDomeTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE);
	m_hCommandList->endTrackingBufferState(m_hRenderSkyVSCB, nvrhi::ResourceStates::CONSTANT_BUFFER);
	m_hCommandList->endTrackingBufferState(m_hPSCB, nvrhi::ResourceStates::CONSTANT_BUFFER);
	m_hCommandList->close();
	m_pDevice->executeCommandList(m_hCommandList);

	return true;
}

void SkyRenderer::BackBufferResizing()
{
	m_hRenderSkyFBs.clear();
	m_pState->hDepthBufferTexture = nullptr;
}

void SkyRenderer::BackBufferResized(const uint32_t width, const uint32_t height, const uint32_t sampleCount)
{
	m_backBufferWidth = width;
	m_backBufferHeight = height;
	m_sampleCount = sampleCount;
	
	if (m_pDevice == nullptr) return;
	if (m_pState == nullptr) return;

	m_needToRecreatePSOs = true;
}


void SkyRenderer::Render(nvrhi::ITexture* renderTarget, uint32_t swapChainIndex)
{
	if (m_pState->bRenderSky)
	{
		m_hCommandList->open();
		nvrhi::FramebufferHandle FB = getFramebuffer(renderTarget, swapChainIndex);
		m_hCommandList->beginTrackingTextureState(m_pState->hDynamicSkyDomeTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE);
		m_hCommandList->beginTrackingBufferState(m_hRenderSkyVSCB, nvrhi::ResourceStates::CONSTANT_BUFFER);
		m_hCommandList->beginTrackingBufferState(m_hPSCB, nvrhi::ResourceStates::CONSTANT_BUFFER);
		m_hCommandList->beginTrackingTextureState(renderTarget, nvrhi::AllSubresources, nvrhi::ResourceStates::RENDER_TARGET);
		m_hCommandList->beginTrackingTextureState(m_pState->hDepthBufferTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::DEPTH_WRITE);

		// Initialize sky draw state
		nvrhi::Viewport vp;
		vp.maxX = float(m_backBufferWidth);
		vp.maxY = float(m_backBufferHeight);

		nvrhi::Rect scissor;
		scissor.maxX = m_backBufferWidth;
		scissor.maxY = m_backBufferHeight;
	
		m_drawState.viewport.viewports.resize(1);
		m_drawState.viewport.viewports[0] = vp;
		m_drawState.viewport.scissorRects = { scissor };

		// Update constant buffers
		vs_sky_cbuffer VSCB;

		gfsdk_float4x4 viewProjMatrix = m_pState->Camera.GetWorldToClip();
		gfsdk_float3 eyePoint = m_pState->Camera.GetCameraPosition();

		VSCB.g_matViewProj = { viewProjMatrix._11, viewProjMatrix._12, viewProjMatrix._13, viewProjMatrix._14,
								viewProjMatrix._31, viewProjMatrix._32, viewProjMatrix._33,viewProjMatrix._34,
								viewProjMatrix._21, viewProjMatrix._22, viewProjMatrix._23,viewProjMatrix._24,
								viewProjMatrix._41, viewProjMatrix._42, viewProjMatrix._43,viewProjMatrix._44 };
		m_hCommandList->writeBuffer(m_hRenderSkyVSCB, &VSCB, sizeof(VSCB));
		
		ps_sky_cbuffer PSCB;
		PSCB.g_sunIntensity = m_pState->fSunIntensity;
		PSCB.g_sunDirection.x = cosf(m_pState->fSunAngle * 0.0174533f);
		PSCB.g_sunDirection.y = 0;
		PSCB.g_sunDirection.z = sinf(m_pState->fSunAngle * 0.0174533f);
		PSCB.g_useDynamicSkyTexture = m_pState->bUseDynamicSky ? 1.0f : 0.0f;
		m_hCommandList->writeBuffer(m_hPSCB, &PSCB, sizeof(PSCB));

		// Rendering
		nvrhi::DrawArguments da;
		da.vertexCount = SKYDOME_SLICES_COUNT*(SKYDOME_SLICES_COUNT + 2) * 2;
		da.startIndexLocation = 0;
		da.startVertexLocation = 0;

		m_drawState.indexBuffer.handle = m_hRenderSkyIndexBuffer;
		m_drawState.indexBuffer.format = nvrhi::Format::R32_UINT;
		m_drawState.indexBuffer.offset = 0;
		m_drawState.vertexBuffers.resize(1);
		m_drawState.vertexBuffers[0].buffer = m_hRenderSkyVertexBuffer;
		m_drawState.vertexBuffers[0].slot = 0;
		m_drawState.vertexBuffers[0].offset = 0;

		m_drawState.framebuffer = FB;
		m_drawState.pipeline = m_hRenderSkyPSO;
		m_drawState.bindings = { m_hRenderSkyBindingSet };

		m_hCommandList->setGraphicsState(m_drawState);
		m_hCommandList->drawIndexed(da);

		m_hCommandList->endTrackingTextureState(m_pState->hDynamicSkyDomeTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE);
		m_hCommandList->endTrackingBufferState(m_hRenderSkyVSCB, nvrhi::ResourceStates::CONSTANT_BUFFER);
		m_hCommandList->endTrackingBufferState(m_hPSCB, nvrhi::ResourceStates::CONSTANT_BUFFER);
		m_hCommandList->endTrackingTextureState(renderTarget, nvrhi::AllSubresources, nvrhi::ResourceStates::RENDER_TARGET);
		m_hCommandList->endTrackingTextureState(m_pState->hDepthBufferTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::DEPTH_WRITE);
		m_hCommandList->close();
		m_pDevice->executeCommandList(m_hCommandList);
	}
}

nvrhi::FramebufferHandle SkyRenderer::getFramebuffer(nvrhi::TextureHandle renderTarget, uint32_t swapChainIndex)
{
	nvrhi::TextureDesc td = renderTarget->GetDesc();

	// Reallocate the depth buffer if needed
	if (m_pState->hDepthBufferTexture == nullptr)
	{
		nvrhi::TextureDesc desc;
		desc.width = td.width;
		desc.height = td.height;
		desc.sampleCount = td.sampleCount;
		desc.debugName = "Depth target created in OceanRenderer::getFramebuffer";
		desc.format = nvrhi::Format::D24S8;
		desc.isRenderTarget = true;
		desc.clearValue = nvrhi::Color(1.0f);
		desc.useClearValue = true;
		m_pState->hDepthBufferTexture = m_pDevice->createTexture(desc);
		assert(m_pState->hDepthBufferTexture);
		m_hCommandList->beginTrackingTextureState(m_pState->hDepthBufferTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::COMMON);
		m_hCommandList->endTrackingTextureState(m_pState->hDepthBufferTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::DEPTH_WRITE);
	}

	// Create framebuffers if needed
	if (swapChainIndex >= m_hRenderSkyFBs.size()) m_hRenderSkyFBs.resize(swapChainIndex + 1);

	if (m_hRenderSkyFBs[swapChainIndex] == nullptr)
	{
		nvrhi::FramebufferDesc desc = nvrhi::FramebufferDesc()
			.addColorAttachment(renderTarget)
			.setDepthAttachment(m_pState->hDepthBufferTexture);

		m_hRenderSkyFBs[swapChainIndex] = m_pDevice->createFramebuffer(desc);
	}


	if (m_needToRecreatePSOs)
	{
		// Recreating PSOs
		nvrhi::GraphicsPipelineDesc renderSkyPSOdesc = m_hRenderSkyPSO->GetDesc();
		m_hRenderSkyPSO = nullptr;
		m_hRenderSkyPSO = m_pDevice->createGraphicsPipeline(renderSkyPSOdesc, m_hRenderSkyFBs[swapChainIndex]);
		m_needToRecreatePSOs = false;
	}

	assert(m_hRenderSkyFBs.at(swapChainIndex));
	return m_hRenderSkyFBs[swapChainIndex];
}

void SkyRenderer::Shutdown()
{
	m_pState->hDynamicSkyDomeTexture = nullptr;
}

bool SkyRenderer::KeyboardUpdate(int key, int scancode, int action, int mods)
{
	m_pState->Camera.KeyboardUpdate(key, scancode, action, mods);
	return false;
}

bool SkyRenderer::MousePosUpdate(double xpos, double ypos)
{
	m_pState->Camera.MousePosUpdate(xpos, ypos);
	return false;
}

bool SkyRenderer::MouseButtonUpdate(int button, int action, int mods)
{
	m_pState->Camera.MouseButtonUpdate(button, action, mods);
	return false;
}

void SkyRenderer::Animate(float elapsedTimeSeconds)
{
	// update dynamic sky texture if/when needed
	if (m_pState->bNeedToUpdateSky)
	{
		m_pState->bNeedToUpdateSky = false;

		m_hCommandList->open();
		m_hCommandList->beginTrackingTextureState(m_hStaticSkyDomeTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE);
		m_hCommandList->beginTrackingTextureState(m_pState->hDynamicSkyDomeTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE);
		m_hCommandList->beginTrackingBufferState(m_hPSCB, nvrhi::ResourceStates::CONSTANT_BUFFER);

		// Initialize sky draw state
		nvrhi::Viewport vp;
		vp.maxX = float(SKYDOME_TEXTURE_SIZE);
		vp.maxY = float(SKYDOME_TEXTURE_SIZE);

		nvrhi::Rect scissor;
		scissor.maxX = SKYDOME_TEXTURE_SIZE;
		scissor.maxY = SKYDOME_TEXTURE_SIZE;

		m_drawState.viewport.viewports.resize(1);
		m_drawState.viewport.viewports[0] = vp;
		m_drawState.viewport.scissorRects = { scissor };

		// Update constant buffers
		ps_sky_cbuffer PSCB;
		PSCB.g_sunIntensity = m_pState->fSunIntensity;
		PSCB.g_sunDirection.x = cosf(m_pState->fSunAngle * 0.0174533f);
		PSCB.g_sunDirection.y = 0;
		PSCB.g_sunDirection.z = sinf(m_pState->fSunAngle * 0.0174533f);
		PSCB.g_useDynamicSkyTexture = m_pState->bUseDynamicSky ? 1.0f : 0.0f;
		m_hCommandList->writeBuffer(m_hPSCB, &PSCB, sizeof(PSCB));

		nvrhi::DrawArguments da;
		da.vertexCount = 4;
		da.startIndexLocation = 0;
		da.startVertexLocation = 0;

		m_drawState.indexBuffer.handle = m_hUpdateSkyIndexBuffer;
		m_drawState.indexBuffer.format = nvrhi::Format::R32_UINT;
		m_drawState.indexBuffer.offset = 0;
		m_drawState.vertexBuffers.resize(1);
		m_drawState.vertexBuffers[0].buffer = m_hUpdateSkyVertexBuffer;
		m_drawState.vertexBuffers[0].slot = 0;
		m_drawState.vertexBuffers[0].offset = 0;
		m_drawState.pipeline = m_hUpdateSkyPSO;
		m_drawState.bindings = { m_hUpdateSkyBindingSet };
		m_drawState.framebuffer = m_hUpdateSkyFB;

		// Rendering to mip #0
		m_hCommandList->setGraphicsState(m_drawState);
		m_hCommandList->drawIndexed(da);

		// Generatin mipmap chain
		GenerateMipmaps();

		m_hCommandList->endTrackingTextureState(m_hStaticSkyDomeTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE);
		m_hCommandList->endTrackingTextureState(m_pState->hDynamicSkyDomeTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE);
		m_hCommandList->endTrackingBufferState(m_hPSCB, nvrhi::ResourceStates::CONSTANT_BUFFER);
		m_hCommandList->close();
		m_pDevice->executeCommandList(m_hCommandList);
	}
}

void SkyRenderer::GenerateMipmaps()
{
	nvrhi::DrawArguments da;
	da.vertexCount = 4;
	da.startIndexLocation = 0;
	da.startVertexLocation = 0;

	m_drawState.indexBuffer.handle = m_hUpdateSkyIndexBuffer;
	m_drawState.indexBuffer.format = nvrhi::Format::R32_UINT;
	m_drawState.indexBuffer.offset = 0;
	m_drawState.vertexBuffers.resize(1);
	m_drawState.vertexBuffers[0].buffer = m_hUpdateSkyVertexBuffer;
	m_drawState.vertexBuffers[0].slot = 0;
	m_drawState.vertexBuffers[0].offset = 0;

	int source_mip_level_size = SKYDOME_TEXTURE_SIZE;
	for (int mip_level = 0; mip_level < SKYDOME_TEXTURE_MIPMAPCOUNT - 1; mip_level++)
	{
		m_drawState.viewport.viewports.resize(1);
		m_drawState.viewport.viewports[0] = nvrhi::Viewport(float(source_mip_level_size / 2), float(source_mip_level_size / 2));
		m_drawState.viewport.scissorRects.resize(1);
		m_drawState.viewport.scissorRects[0] = nvrhi::Rect(source_mip_level_size / 2, source_mip_level_size / 2);

		m_drawState.pipeline = m_hMipGenPSO[mip_level];
		m_drawState.framebuffer = m_hMipGenFB[mip_level];
		m_drawState.bindings = { m_hMipGenBindingSet[mip_level] };

		m_hCommandList->setGraphicsState(m_drawState);
		m_hCommandList->drawIndexed(da);

		source_mip_level_size /= 2;
	}
}
